#lang racket

;; halts?: (any -> any) any -> boolean
;; Interpretation:
;;   Given a function (fn) and an argument (arg) of the correct types,
;;   returns true if (fn arg) returns a value, or false otherwise
;;   (i.e. an infinite loop or an exception)
(define (halts? fn arg) ...)

;; self-halt?: (fn -> any) fn -> boolean
;; Interpretation:
;;   Given a function (fn), returns true if (fn fn) returns a value, or
;;   false otherwise (i.e. an infinite loop or an exception)
(define (self-halt? fn)
  (halts? fn fn))

;; loop: -> void
;; Interpretation: an infinite loop
(define (loop) (loop))

;; oops: (fn -> any) fn -> boolean
;; Interpretation:
;;   Given a function (fn), returns true if (fn fn) goes into an infinite loop
;;   or throws an exception, and goes into an infinite loop if (fn fn) returns
;;   a value
(define (oops fn)
  (if (self-halt? fn)
      (loop)
      true))

(oops oops)
;; Does this loop or return an answer?
;; (a) it loops.  Then (self-halts? oops) must be true.  But that means
;;     that (halts? (oops oops)) must be true.  But that means that
;;     (oops oops) returns a value, and we said it loops.  WRONG.
;; (b) it returns a value.  Then (self-halts? oops) must be false.  But that
;;     means that (halts? (oops oops)) must be false.  But that means that
;;     (oops oops) must go into a loop, and we said it returns a
;;     value.  WRONG. 

